# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import time
import urwid

last_key_press = None
pending_vim_command = None

def orig_w(widget):
    if isinstance(widget, urwid.AttrMap):
        return widget.original_widget
    return widget

def vim2emacs_translation(key):
    global last_key_press
    global pending_vim_command

    if key == 'j' or key == 'J':
        return 'down'

    if key == 'k' or key == 'K':
        return 'up'

    if key == 'ctrl u':
        return 'page up'

    if key == 'ctrl d':
        return 'page down'

    if key == 'G':
        return 'end'

    if key == 'g' and not pending_vim_command:
        last_key_press = time.time()
        pending_vim_command = 'g'
        return None

    if key == 'g' and pending_vim_command:
        diff = time.time() - last_key_press
        last_key_press = None
        pending_vim_command = None

        if diff < 0.25:
            return 'home'

    return key



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import logging

from mitzasql import constants as const

log_dir = os.path.dirname(const.DEFAULT_LOG_PATH)

os.makedirs(log_dir, mode=0o700, exist_ok=True)

logger = logging.getLogger(const.APP_NAME)
logger.setLevel(logging.DEBUG)

file_handler = logging.FileHandler(const.DEFAULT_LOG_PATH, encoding='utf-8')
file_handler.setLevel(logging.DEBUG)

log_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
file_handler.setFormatter(log_formatter)

logger.addHandler(file_handler)


class LoggerMixin:
    def set_log_prefix(self, prefix):
        self._prefix = prefix

    def _make_msg(self, msg):
        return self._prefix + ':' + msg

    def log_debug(self, msg, *args, **kwargs):
        msg = self._make_msg(msg)
        logger.debug(msg, *args, **kwargs)

    def log_info(self, msg, *args, **kwargs):
        msg = self._make_msg(msg)
        logger.info(msg, *args, **kwargs)

    def log_warning(self, msg, *args, **kwargs):
        msg = self._make_msg(msg)
        logger.warning(msg, *args, **kwargs)

    def log_error(self, msg, *args, **kwargs):
        msg = self._make_msg(msg)
        logger.error(msg, *args, **kwargs)

    def log_critical(self, msg, *args, **kwargs):
        msg = self._make_msg(msg)
        logger.critical(msg, *args, **kwargs)

    def log_exception(self, msg, *args, **kwargs):
        msg = self._make_msg(msg)
        logger.exception(msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        msg = self._make_msg(msg)
        logger.log(level, msg, *args, **kwargs)



import os
import pytest

from mitzasql.db.connection import Connection

db_host = os.getenv('DB_HOST', 'tcp://localhost')
db_port = os.getenv('DB_PORT', '3306')
db_username = os.getenv('DB_USER', 'root')
db_password = os.getenv('DB_PASS', '')

@pytest.fixture(scope='module')
def connection():
    con_data = {
            'host': db_host,
            'port': db_port,
            'username': db_username,
            'password': db_password,
            'database': ''
            }
    con, error = Connection.factory(con_data, 'localhost')
    assert error is None
    return con

@pytest.fixture(scope='module')
def sakila_connection():
    con_data = {
            'host': db_host,
            'port': db_port,
            'username': db_username,
            'password': db_password,
            'database': 'sakila'
            }
    con, error = Connection.factory(con_data, 'localhost')
    assert error is None
    return con



import pytest

from mitzasql.db.model import Model


@pytest.fixture
def basic_dataset():
    return [(1, 2, 3), (4, 5, 6), (7, 8, 9)]

@pytest.fixture
def basic_columns():
    return (
            {
            'name': 'first_no',
            'is_int': 1,
            'flags': None
            },
            {
            'name': 'second_no',
            'is_int': 1,
            'flags': None
            },
            {
            'name': 'third_no',
            'is_int': 1,
            'flags': None
            }
        )

def test_model_returns_correct_rowcount(basic_dataset, basic_columns):
    model = Model(basic_dataset, basic_columns)
    assert len(model) == 3

def test_model_returns_item(basic_dataset, basic_columns):
    model = Model(basic_dataset, basic_columns)
    assert model[1] == (4, 5, 6)

def test_model_returns_correct_column_index(basic_dataset, basic_columns):
    model = Model(basic_dataset, basic_columns)
    assert model.col_index('second_no') == 1

def test_model_searches_for_keyword(basic_dataset, basic_columns):
    model = Model(basic_dataset, basic_columns)
    result = model.search('4', 0)
    assert result is not None

    row_index, row = result
    assert row_index == 1
    assert row == (4, 5, 6)

def test_model_searches_for_keyword_in_reverse(basic_dataset, basic_columns):
    model = Model(basic_dataset, basic_columns)
    result = model.search('4', 0, pos=2, reverse=True)
    assert result is not None

    row_index, row = result
    assert row_index == 1
    assert row == (4, 5, 6)

def test_model_search_returns_none_in_reverse(basic_dataset, basic_columns):
    model = Model(basic_dataset, basic_columns)
    result = model.search('8', 1, pos=1, reverse=True)
    assert result is None



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.ui.widgets.base_db_view import BaseDBView
from mitzasql.ui.widgets.table import Table
from mitzasql.ui.widgets.cmd_proc import (BaseCmdProcessor, SearchCmdProcessor)

class CommandProcessor(BaseCmdProcessor, SearchCmdProcessor):
    def __init__(self, dbview):
        self._dbview = dbview
        BaseCmdProcessor.__init__(self)
        SearchCmdProcessor.__init__(self, self._dbview.search_db)
        self._cmd_strings = {
                '/': self.search
                }
        self._cmd_keys = {
                'n': self.search_next,
                'N': self.search_prev
                }

class DBView(BaseDBView):
    SIGNAL_ACTION_SELECT_DB = 'select_db'
    SIGNAL_ACTION_PROCESS_LIST = 'process_list'

    def __init__(self, model, connection):
        self._command_processor = CommandProcessor(self)
        self._table_widget_cls = Table
        super().__init__(model, connection)
        self.SIGNALS.append(self.SIGNAL_ACTION_SELECT_DB)
        urwid.connect_signal(self._table, self._table.SIGNAL_ROW_SELECTED, self.select_db)

    def __del__(self):
        urwid.disconnect_signal(self._table, self._table.SIGNAL_ROW_SELECTED,
                self.select_db)
        super().__del__()

    def select_db(self, emitter, row):
        database = row[0]
        urwid.emit_signal(self, self.SIGNAL_ACTION_SELECT_DB, self, database)

    def search_db(self, db, pos=0, reverse=False):
        result = self._model.search(db, col_index=0, pos=pos, reverse=reverse)
        if result is None:
            return
        index, row = result
        self._table.focus_row(index)
        return (index, row)



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.ui.widgets.base_db_view import BaseDBView
from mitzasql.ui.widgets.table import Table

from mitzasql.ui.widgets.cmd_proc import (BaseCmdProcessor, SearchCmdProcessor, CommandError)

class CommandProcessor(BaseCmdProcessor):
    def __init__(self, query_view):
        self._query_view = query_view
        BaseCmdProcessor.__init__(self)

        self._colon_cmds.extend([
                ('r', 'resize', self._resize_table)
                ])

        self._cmd_strings = {
                ':': self.colon_handler
                }

    def _resize_table(self, *args):
        if not len(args):
            return
        if len(args) == 1:
            column = args[0]
            value = 1
        else:
            column, value = args
        self._query_view.resize_tbl_col(column, value)

class QueryView(BaseDBView):
    SIGNAL_ACTION_SELECT_ROW = 'select_row'

    def __init__(self, model, connection):
        self._command_processor = CommandProcessor(self)
        self._table_widget_cls = Table
        super().__init__(model, connection)
        self._command_processor.cmd_args_suggestions['resize'] = [c['name'] for c in self._model.columns]
        if connection.database is None:
            database = ''
        else:
            database = connection.database
        self.SIGNALS.append(self.SIGNAL_ACTION_SELECT_ROW)
        self._connect_table_signals()

    def _connect_table_signals(self):
        urwid.connect_signal(self._table, self._table.SIGNAL_ROW_SELECTED,
                self.select_row)

    def _disconnect_table_signals(self):
        urwid.disconnect_signal(self._table, self._table.SIGNAL_ROW_SELECTED,
                self.select_row)

    def select_row(self, emitter, row):
        urwid.emit_signal(self, self.SIGNAL_ACTION_SELECT_ROW, self, row)

    def refresh(self, query, connection):
        if query == self._model.query:
            return

        self._model.query = query
        self._model.reload()
        self._query_editor.query = query
        self._update_breadcrumbs()

    def resize_tbl_col(self, column, value=1):
        try:
            value = int(value)
            col_index = self._model.col_index(column.lower())
        except IndexError as e:
            raise CommandError(u'Column not found!')
        except ValueError as e:
            raise CommandError(u'Invalid column width!')
        self._table.resize_col(col_index, value)



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

class EmacsKeysMixin:
    def keypress(self, size, key):
        if key == 'ctrl e':
            return 'end'
        if key == 'ctrl a':
            return 'home'
        if key == 'ctrl h':
            return 'backspace'
        if key == 'ctrl w':
            self._remove_word(direction='back')
            return
        if key == 'meta d':
            self._remove_word(direction='forward')
            return
        if key == 'meta b':
            self._move_to_word(direction='back')
            return
        if key == 'meta f':
            self._move_to_word(direction='forward')
            return
        if key == 'ctrl f':
            self.edit_pos += 1
            return
        if key == 'ctrl b':
            if self.edit_pos != 0:
                self.edit_pos -= 1
            return
        return key

    def _remove_word(self, direction):
        '''Remove word until first separator character (back or forward)'''
        old_pos = self.edit_pos
        self._move_to_word(direction)
        new_pos = self.edit_pos
        if direction == 'back':
            self.edit_text = self.edit_text[0:new_pos] + self.edit_text[old_pos:]
            self.edit_pos = new_pos
        else:
            self.edit_text = self.edit_text[0:old_pos] + self.edit_text[new_pos:]
            self.edit_pos = old_pos


    def _move_to_word(self, direction):
        if direction == 'back':
            if self.edit_pos == 0:
                return
            text = self.edit_text[0:self.edit_pos-1]
            pos = self.edit_pos - 1
            found_alnum = False
            for char in reversed(text):
                pos -= 1
                if char.isalnum():
                    self.edit_pos = pos
                    found_alnum = True
                    continue
                if found_alnum:
                    return

            if not found_alnum:
                self.edit_pos = 0
            return

        if self.edit_pos == len(self.edit_text):
            return

        text = self.edit_text[self.edit_pos+1:]
        pos = self.edit_pos
        found_sep = False
        for char in text:
            pos += 1
            if not char.isalnum():
                self.edit_pos = pos
                found_sep = True
            elif found_sep is True:
                return

        if not found_sep:
            self.edit_pos = len(self.edit_text)



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.ui.widgets.emacs_edit import EmacsEdit
from mitzasql.ui.widgets.list import List

class TableChangerWidget(urwid.Frame):
    SIGNAL_ESCAPE = 'escape'
    SIGNAL_CHANGE_TABLE = 'change_table'

    def __init__(self, model):
        self._model = model;
        self._search_input = EmacsEdit()
        self._tables_list = self._create_tables_list()
        header = urwid.Pile([urwid.AttrMap(self._search_input, 'editbox'), urwid.Divider('-')])
        super().__init__(self._tables_list, header, focus_part='header')
        urwid.register_signal(self.__class__, [self.SIGNAL_ESCAPE, self.SIGNAL_CHANGE_TABLE])
        urwid.connect_signal(self._search_input, 'change', self.filter_list)
        urwid.connect_signal(self._tables_list,
                self._tables_list.SIGNAL_SELECTED, self.on_select_table)

    def on_select_table(self, emitter, value):
        urwid.emit_signal(self, self.SIGNAL_CHANGE_TABLE, self, value)

    def filter_list(self, emitter, value):
        self._tables_list.clear()
        for item in self._model:
            if not self._is_table(item):
                continue
            name = item[0]
            if not name.lower().startswith(value):
                continue
            self._tables_list.append(name)

    def keypress(self, size, key):
        if key == 'esc':
            urwid.emit_signal(self, self.SIGNAL_ESCAPE, self)
            return

        if key == 'tab':
            if self.focus_position == 'body':
                self.focus_position = 'header'
            else:
                self.focus_position = 'body'
            return

        if key == 'enter' and self.focus_position == 'header' and len(self._tables_list.body):
            self.focus_position = 'body'

        return super().keypress(size, key)

    def _is_table(self, row):
        type = row[-1]
        if type != 'BASE TABLE' and type != 'VIEW' and type != 'SYSTEM VIEW':
            return False
        return True

    def _create_tables_list(self):
        tables = []
        for item in self._model:
            if not self._is_table(item):
                continue
            tables.append(item[0])

        tables_list = List(tables)
        return tables_list




# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import pickle
import hashlib
import urwid
import time
import sys

from mitzasql.ui import utils
from mitzasql.logger import logger

class InputProcessor():
    '''
    Implements support for VIM style keys & basic commands.

    Supported keys:
        j,k - scroll 1 row up/down
        h,l - scroll 1 column left/right
        ctrl-[u/d] - scroll rows page up/down
        0 - go to first column
        $ - go to last column
        G - go to last row

    Supported commands:
        [number]j/k/h/l - scroll [number] rows/columns
        gg - go to first row
    '''
    def __init__(self):
        self._last_key_press = None
        self._pending_command = None

        self._keys_direction = {
                'home': 'up',
                'ctrl u': 'up',
                'up' : 'up',
                'k': 'up',
                'page up': 'up',

                'end': 'down',
                'down': 'down',
                'j': 'down',
                'ctrl d': 'down',
                'G': 'down',
                'page down': 'down',

                'left': 'left',
                '0': 'left',
                'h': 'left',
                'ctrl left': 'left',

                'right': 'right',
                '$': 'right',
                'l': 'right',
                'ctrl right': 'right'
                }

        self._single_offset_keys = ['up', 'down', 'left', 'right', 'j', 'k', 'h', 'l']
        self._max_offset_keys = ['home', 'end', 'ctrl left', 'ctrl right', '0', '$', 'G']
        self._page_offset_keys = ['page up', 'page down', 'ctrl u', 'ctrl d']

        self._digit_keys = [str(i) for i in range(1, 10)]
        self._command_keys = ['g']
        self._command_keys.extend(self._digit_keys)

    def process_key(self, key):
        if not self._pending_command:
            if key in self._single_offset_keys:
                return (self._keys_direction[key], 1)

            if key in self._max_offset_keys:
                direction = self._keys_direction[key]
                return (direction, sys.maxsize)

            if key in self._page_offset_keys:
                direction = self._keys_direction[key]
                return (direction, 'page')

            if key in self._command_keys:
                self._last_key_press = time.time()
                self._pending_command = key
                return (None, None)

            return (None, None)

        diff = time.time() - self._last_key_press
        if diff > self._timeout(key):
            self._clear_pending_command()
            return (None, None)

        command = self._process_command(key)
        return (None, None, command)


    def _timeout(self, key):
        if self._pending_command.isdigit():
            return 1
        return 0.25

    def _clear_pending_command(self):
        self._pending_command = None
        self._last_key_press = None

    def _process_command(self, key):
        if key == 'esc':
            self._clear_pending_command()
            return

        if self._pending_command.isdigit() and key.isdigit():
            self._pending_command += key
            return

        if self._pending_command.isdigit() and not key.isdigit():
            offset = int(self._pending_command)
            self._clear_pending_command()

            if not key in self._keys_direction:
                return
            return ('scroll', self._keys_direction[key], offset)


        if self._pending_command == 'g' and key == 'g':
            self._clear_pending_command()
            return ('scroll', 'up', sys.maxsize)

class TCell(urwid.Text):
    def __init__(self, content, width, separator=True, align='left', separator_attr=None):
        self.separator = separator
        self._content = content
        self._align = align
        self._width = width
        self._separator_attr = separator_attr

        content = self._format_content()
        super().__init__(content, 'left', 'clip')

    def _format_content(self):
        content_attr = None
        if not isinstance(self._content, tuple):
            content = str(self._content)
        else:
            content_attr = self._content[0]
            content = str(self._content[1])

        if self.separator is True:
            content_max_len = self._width - 1
        else:
            content_max_len = self._width

        if len(content) > content_max_len:
            content = content[0:content_max_len - 1]
            content += chr(187)
        else:
            if self._align == 'left':
                content = content.ljust(content_max_len)
            else:
                content = content.rjust(content_max_len)

        if content_attr is not None:
            content = (content_attr, content)

        if self.separator is True:
            if self._separator_attr is not None:
                sep = (self._separator_attr, chr(124))
            else:
                sep = chr(124)
            if isinstance(content, list):
                content.append(sep)
            else:
                content = [content, sep]
        return content

    def resize(self, width):
        self._width = width
        content = self._format_content()
        self.set_text(content)

class TCellText(TCell):
    def __init__(self, content, width, separator=True):
        if isinstance(content, set):
            content = str(content)
        content = content.replace("\n", "\\n")
        super().__init__(content, width, separator)

class TCellNumber(TCell):
    def __init__(self, content, width, separator=True):
        super().__init__(content, width, separator, align='right')

class TCellNull(TCell):
    def __init__(self, width, column_info, separator=True):
        if column_info['is_int'] is True or column_info['is_real'] is True:
            align = 'right'
        else:
            align = 'left'
        content = ('tcell_null_unfocused', u'(NULL)')
        super().__init__(content, width, separator, align=align)

class TCellBinary(TCell):
    def __init__(self, content, width, separator=True):
        if not isinstance(content, str):
            try:
                content = content.decode(encoding='utf8')
            except:
                content = content.hex();

        if isinstance(content, str):
            content = content.replace("\n", "\\n")
        super().__init__(content, width, separator)

class TCellSpatial(TCell):
    def __init__(self, content, width, separator=True):
        if isinstance(content, str):
            content = content.replace("\n", "\\n")
        elif isinstance(content, bytearray):
            content = content.hex();
        super().__init__(content, width, separator)


class TRow(urwid.Columns):
    MIN_COL_LENGTH = 3
    MAX_COL_LENGTH = 80

    def __init__(self, cols, selectable=True):
        self._selectable = selectable
        super().__init__(cols, dividechars=1)

    def selectable(self):
        return self._selectable

    def keypress(self, size, key):
        return key

    def resize(self, col_index, increment):
        cell, options = self.contents[col_index]

        width = int(options[1]) + increment
        if width < self.MIN_COL_LENGTH or width > self.MAX_COL_LENGTH:
            return
        cell.resize(width)

        options = self.options('given', width, False)
        self.contents[col_index] = (cell, options)
        return width

    def render(self, size, focus=False):
        for col in self.contents:
            if not isinstance(col[0], TCellNull):
                continue
            text, attr = col[0].get_text()
            if focus is True:
                attr = 'tcell_null_focused'
            else:
                attr = 'tcell_null_unfocused'

            if col[0].separator is True:
                text = [(attr, text[0:-1]), text[-1:]]
            else:
                text = (attr, text)
            col[0].set_text(text)
        return super().render(size, focus)

class THeader(TRow):
    def __init__(self, columns):
        self.cell_sizes = []
        self._cols_hash = self._get_cols_hash(columns)
        cols = self._create_cols(columns)
        super().__init__(cols, False)

    def _get_cols_hash(self, columns):
        serialized = pickle.dumps(columns)
        m = hashlib.sha256()
        m.update(serialized)
        hsh = m.digest()
        return hsh

    def _create_cols(self, columns):
        cols = []
        index = 0
        columns_count = len(columns)

        for info in columns:
            name = info['name']
            size = info['max_len']
            self.cell_sizes.append(size)

            if index < columns_count - 1:
                separator = True
            else:
                separator = False

            if info['is_int'] is True or info['is_real'] is True:
                align = 'right'
            else:
                align = 'left'

            cell = TCell(name, size, separator, align=align, separator_attr='theader_sep')
            cols.append((size, cell))
            index += 1
        return cols

    def refresh(self, columns):
        cols_hash = self._get_cols_hash(columns)
        if cols_hash == self._cols_hash:
            return
        self._cols_hash = cols_hash
        self.cell_sizes = []
        self.contents.clear()
        cols = self._create_cols(columns)
        for width, widget in cols:
            col_size = self.options(width_type='given', width_amount=width)
            self.contents.append((widget, col_size))

    def resize(self, col_index, increment):
        width = super().resize(col_index, increment)
        if width is not None:
            self.cell_sizes[col_index] = width
        return width

class TBody(urwid.ListBox):
    KEYPRESS = 'keypress'

    def __init__(self, columns_list, cell_sizes, rows=[]):
        self._columns_list = columns_list
        self._cell_sizes = cell_sizes

        urwid.register_signal(self.__class__, self.KEYPRESS)
        super().__init__(urwid.SimpleFocusListWalker(rows))

    def set_columns(self, columns_list, cell_sizes):
        self._columns_list = columns_list
        self._cell_sizes = cell_sizes

    def update_col_width(self, col_index, width):
        self._cell_sizes[col_index] = width

    def clear(self):
        self.body.clear()

    def add_row(self, row_data):
        row = self.make_row(row_data)
        self.body.append(row)

    def __iter__(self):
        for row in self.body:
            yield row

    @property
    def focused_row(self):
        row, index = self.body.get_focus()
        return row.original_widget

    def make_row(self, columns):
        cols = []
        index = 0
        columns_count = len(columns)

        for index, value in enumerate(columns):
            column_info = self._columns_list[index]
            cell_size = self._cell_sizes[index]

            if index < columns_count - 1:
                separator = True
            else:
                separator = False

            cell = self._make_cell(value, cell_size, separator, column_info)

            cols.append((cell_size, cell))
        row = urwid.AttrMap(TRow(cols), 'default', focus_map='trow_focused')
        return row

    def _make_cell(self, value, size, separator, column_info):
        if value is None:
            cell = TCellNull(size, column_info, separator)
        elif column_info['is_binary'] is True:
            cell = TCellBinary(value, size, separator)
        elif column_info['is_int'] is True or column_info['is_real'] is True:
            cell = TCellNumber(value, size, separator)
        elif column_info['is_text'] is True:
            cell = TCellText(value, size, separator)
        elif column_info['is_spatial'] is True:
            cell = TCellSpatial(value, size, separator)
        else:
            cell = TCell(value, size, separator)
        return cell

    def keypress(self, size, key):
        '''Emit a signal with the key and size to be handled in the parent

        The reason for this is to get the inner size of the table
        Disable only the 'up' and 'down' keys due to interfering
        with other parent pile containers
        '''
        urwid.emit_signal(self, self.KEYPRESS, self, size, key)
        if key == 'up' or key == 'down':
            return
        return key

class Table(urwid.Frame):
    SIGNAL_ROW_SELECTED = 'selected'
    SIGNALS = [SIGNAL_ROW_SELECTED]

    def __init__(self, model):
        self._model = model
        self._rowcount = len(self._model)
        self._header = THeader(self._model.columns)
        self._body = self.make_body()
        self._footer = urwid.Text('')
        self._widget_size = None
        self._visible_columns = 0
        self._focused_row_index = 0
        self._focused_col_index = 0
        self._input_processor = InputProcessor()

        urwid.connect_signal(model, model.SIGNAL_LOAD, self.refresh)

        super().__init__(self._body,
                urwid.AttrMap(self._header, 'theader'),
                urwid.AttrMap(self._footer, 'tfooter'))
        urwid.register_signal(self.__class__, self.SIGNALS)
        self._update_footer()

    def __del__(self):
        urwid.disconnect_signal(self._model, self._model.SIGNAL_LOAD, self.refresh)
        urwid.disconnect_signal(self._body, self._body.KEYPRESS, self.on_body_keypress)
        self._body.clear()

    def focus_row(self, row_index):
        self._focused_row_index = row_index
        self._scroll_rows()

    def refresh(self, model):
        self._visible_columns = 0
        self._focused_row_index = 0
        self._focused_col_index = 0
        self._rowcount = len(self._model)
        utils.orig_w(self._header).refresh(self._model.columns)
        self._body.clear()
        self._body.set_columns(self._model.columns, utils.orig_w(self._header).cell_sizes)
        for row in self._model:
            self._body.add_row(row)
        self._update_footer()
        self._scroll_rows()

    def resize_col(self, col_index, increment=1):
        width = self._header.original_widget.resize(col_index, increment)

        if width is not None:
            self._body.update_col_width(col_index, width)

        for row in self._body.body:
            trow = row.original_widget
            trow.resize(col_index, increment)

    def _increment_col_index(self, offset, absolute=False):
        if absolute is False:
            new_index = self._focused_col_index + offset
        else:
            new_index = offset

        if new_index < 0:
            new_index = 0
        elif new_index < self._visible_columns - 1:
            new_index = self._visible_columns - 1

        if new_index > (len(self._model.columns) - 1):
            new_index = len(self._model.columns) - 1

        self._focused_col_index = new_index

    def _increment_row_index(self, offset, absolute=False):
        if absolute is False:
            new_index = self._focused_row_index + offset
        else:
            new_index = offset

        if new_index < 0:
            new_index = 0

        if new_index > (self._rowcount - 1):
            new_index = self._rowcount - 1

        self._focused_row_index = new_index

    def make_body(self):
        body = TBody(columns_list=self._model.columns,
                cell_sizes=self._header.cell_sizes)

        urwid.connect_signal(body, body.KEYPRESS, self.on_body_keypress)
        for row in self._model:
            body.add_row(row)
        return body

    def on_body_keypress(self, emitter, size, key):
        self._widget_size = size

        if not self._rowcount:
            return

        self._detect_visible_columns_count()

        row, index = self._body.body.get_focus()
        self._focused_row_index = index

        if key == 'enter':
            urwid.emit_signal(self, self.SIGNAL_ROW_SELECTED, self,
                    self._model[index])

        direction, offset, *command  = self._input_processor.process_key(key)

        if direction is None and len(command) and command[0] is None:
            return

        if len(command):
            command = command.pop()
            if isinstance(command, tuple):
                if command[0] == 'scroll':
                    direction, offset = command[1:]
                else:
                    raise RuntimeError('Unimplemented command {0}'.format(command[0]))

        if offset == 'page':
            offset = size[1] - 2

        if direction == 'down':
            self._scroll_down(offset)
        elif direction == 'up':
            self._scroll_up(offset)
        elif direction == 'left':
            self._scroll_left(offset)
        elif direction == 'right':
            self._scroll_right(offset)

        self._update_footer()

    def _update_footer(self):
        rowcount = self._rowcount
        cols_count = len(self._model.columns)

        row_index = self._focused_row_index

        status = u'[{0}/{1}:{2}]'.format(row_index + 1, rowcount, cols_count)

        self._footer.original_widget.set_text(status)

    def _detect_visible_columns_count(self):
        focused_row = self._body.focused_row
        column_widths = focused_row.column_widths(self._widget_size)
        visible_columns = 0
        for width in column_widths:
            if width > 0:
                visible_columns += 1
        self._visible_columns = visible_columns

    def _scroll_down(self, offset):
        self._increment_row_index(offset)
        self._scroll_rows()

    def _scroll_up(self, offset):
        self._increment_row_index(-1 * offset)
        self._scroll_rows()

    def _scroll_left(self, offset):
        if self._focused_col_index < self._visible_columns - 1:
            self._focused_col_index = self._visible_columns - 1
        else:
            self._increment_col_index(-1 * offset)

        self._scroll_columns()

    def _scroll_right(self, offset):
        if self._focused_col_index == 0:
            self._increment_col_index(self._visible_columns - 1 + offset,
                    absolute=True)
        else:
            self._increment_col_index(1 * offset)

        self._scroll_columns()

    def _scroll_columns(self):
        for row in self._body.body:
            row.original_widget.focus_position = self._focused_col_index
        self._header.original_widget.focus_position = self._focused_col_index

    def _scroll_rows(self):
        if self._focused_row_index >= len(self._model):
            return
        self._body.set_focus(self._focused_row_index)
        # Enforce the column focus because sometimes it resets to 0 and screws up
        # the rendering
        self._scroll_columns()



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.ui.widgets.info_widget import InfoWidget

class RowWidget(InfoWidget):
    def __init__(self, row, columns):
        self._row = row;
        self._columns = columns
        contents = self._create_contents()
        super().__init__([contents])

    @property
    def name(self):
        return u'Row data'

    def _create_contents(self):
        grid = []

        index = 0
        for item in self._row:
            if item is None:
                cell_data = u'NULL'
            else:
                if isinstance(item, bytearray):
                    try:
                        cell_data = item.decode(encoding='utf8')
                    except:
                        cell_data = item.hex()
                else:
                    cell_data = str(item)
            cell_name = str(self._columns[index]['name'])

            contents = []
            contents.append((40, urwid.AttrMap(urwid.Text(cell_name), 'editbox:label')))
            contents.append(urwid.AttrMap(urwid.Text(cell_data),
                'editbox'))
            grid.append(urwid.Columns(contents))
            index += 1

        grid = urwid.GridFlow(grid, cell_width=80, h_sep=1, v_sep=1,
                align='left')
        return grid



import pytest
import urwid

from mitzasql.ui.widgets.db_tables_view import DBTablesView
from mitzasql.db.model import DBTablesModel
from ..db.connection_fixture import sakila_connection

def test_keypresses(sakila_connection):
    model = DBTablesModel(sakila_connection, 'sakila')
    view = DBTablesView(model, sakila_connection)

    size = (80, 30)

    emitter = None
    table = None
    def test_table_is_selected(emitter_, table_):
        nonlocal emitter
        nonlocal table
        emitter = emitter_
        table = table_

    urwid.connect_signal(view, view.SIGNAL_ACTION_SELECT_TABLE,
            test_table_is_selected)

    view.keypress(size, 'down')
    view.keypress(size, 'down')
    view.keypress(size, 'down')
    view.keypress(size, 'up')
    view.keypress(size, 'up')
    view.keypress(size, 'up')

    view.keypress(size, 'enter')

    assert isinstance(emitter, DBTablesView)
    assert table is not None
    old_table = table

    view.keypress(size, 'down')
    view.keypress(size, 'enter')

    assert table is not None
    assert table != old_table

    initial_focus = view._table._focused_row_index
    assert initial_focus > 0

    view.keypress(size, 'f5')
    new_focus = view._table._focused_row_index

    assert new_focus == 0

def test_db_tables_view_searches_for_table(sakila_connection):
    model = DBTablesModel(sakila_connection, 'sakila')
    view = DBTablesView(model, sakila_connection)

    size = (80, 30)

    result = view.search_tables('film')
    assert result is not None

    index, row = result
    assert index > 0
    assert 'del_film' in row

    result = view.search_tables('non existent table')
    assert result is None

def test_resizing_column_doesnt_raise_exception(sakila_connection):
    model = DBTablesModel(sakila_connection, 'sakila')
    view = DBTablesView(model, sakila_connection)

    view.resize_tbl_col('engine', 10)
    view.render((80, 30))



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import threading

import urwid

from mitzasql.logger import logger
from mitzasql.ui import main_loop as shared_main_loop
import mitzasql.ui.utils as utils
from mitzasql.ui.widgets.action_bar import ActionBar
from mitzasql.ui.widgets.command_edit import CommandEdit

class DBViewFooter(urwid.WidgetPlaceholder):
    SIGNAL_EXIT_COMMAND_MODE = 'exit_command_mode'
    def __init__(self, actions, command_processor=None):
        self._action_bar = urwid.AttrMap(ActionBar(actions), 'action_bar')
        self._message_bar = urwid.AttrMap(urwid.Text(''),
                'info_message')
        self._command_bar = None
        self._command_processor = None
        self._clear_error_timer = None

        if command_processor is not None:
            self._command_processor = command_processor
            command_bar = CommandEdit(command_processor)
            self._command_bar = urwid.AttrMap(command_bar,
                    'command_bar')
            urwid.connect_signal(command_bar, command_bar.SIGNAL_CANCEL, self.cancel_command_edit)
            urwid.connect_signal(command_processor, command_processor.SIGNAL_ERROR,
                    self.show_command_error)

        super().__init__(self._action_bar)
        urwid.register_signal(self.__class__, self.SIGNAL_EXIT_COMMAND_MODE)

    def __del__(self):
        self._cancel_clear_error_timer()
        if not self._command_bar:
            return
        command_bar = utils.orig_w(self._command_bar)
        urwid.disconnect_signal(command_bar, command_bar.SIGNAL_CANCEL, self.cancel_command_edit)
        urwid.disconnect_signal(self._command_processor,
                self._command_processor.SIGNAL_ERROR, self.show_command_error)

    def _cancel_clear_error_timer(self):
        if self._clear_error_timer is not None:
            self._clear_error_timer.cancel()

    def show_command_error(self, emitter, error_message, *args):
        self.toggle_message_bar(True, message=error_message, is_error=True)

        self._cancel_clear_error_timer()
        def clear_message():
            self.toggle_message_bar(False)

        self._clear_error_timer = threading.Timer(1, clear_message)
        self._clear_error_timer.daemon = True
        self._clear_error_timer.start()

    def toggle_message_bar(self, status, message=None, is_info=True,
            is_error=False):
        if status is False:
            self.original_widget = self._action_bar
            shared_main_loop.refresh()
            return

        self.original_widget = self._message_bar
        if message is not None:
            utils.orig_w(self._message_bar).set_text(message)
            if is_error is True:
                self._message_bar.set_attr_map({
                    None: 'error_message'
                    })
            else:
                self._message_bar.set_attr_map({
                    None: 'info_message'
                    })
        shared_main_loop.refresh()

    def cancel_command_edit(self, emiter):
        self.toggle_command_mode(False)
        urwid.emit_signal(self, self.SIGNAL_EXIT_COMMAND_MODE, self)

    def toggle_command_mode(self, status, command_marker=None,
            show_last_cmd=False):
        if status is True:
            self._cancel_clear_error_timer()
            self.original_widget = self._command_bar
            if show_last_cmd is True:
                utils.orig_w(self._command_bar).show_last_cmd()
            utils.orig_w(self._command_bar).set_marker(command_marker)
            return
        utils.orig_w(self._command_bar).reset()
        self.original_widget = self._action_bar




import pytest

import urwid
from mitzasql.db.model import DatabasesModel
from .connection_fixture import connection

def test_model_fetches_data(connection):
    model = DatabasesModel(connection)
    assert len(model) > 0
    assert 'Database' in [c['name'] for c in model.columns]

    model_was_iterated = False
    rows = []
    for row in model:
        rows.append(row[0])
        model_was_iterated = True

    assert 'sakila' in rows
    assert len(model.columns)

def test_model_searches_for_sakila(connection):
    model = DatabasesModel(connection)

    result = model.search('saki', 0)
    assert result is not None

    row_index, row = result
    assert row_index > 0
    assert row == ('sakila', )



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

class ActionBar(urwid.Columns):
    """Column widget containing actions labels and associated keys

    It's a cheap clone of the Midnight Commander action status bar
    """

    def __init__(self, actions):
        self._actions = actions
        column_widgets = self._create_widgets()
        super().__init__(column_widgets, dividechars=1)

    def _create_widgets(self):
        widgets = []
        for action_key, label in self._actions:
            label_width = len(label) + 4
            label_markup = self._create_label_markup(action_key, label)
            text = urwid.Padding(urwid.Text(label_markup, align='center'), align='center',
                    width=label_width)
            text = urwid.AttrMap(text, 'action_bar:action')
            widgets.append((label_width, text))
        return widgets

    def _create_label_markup(self, action_key, label):
        key_pos = label.find(action_key)
        if key_pos == -1:
            return label

        label_first_part = label[:key_pos]
        label_middle_part = ('action_bar:action_key', action_key)
        label_last_part = label[key_pos + len(action_key):]

        return [label_first_part, label_middle_part, label_last_part]



import pytest

from mitzasql.db.model import TableInfoModel
from .connection_fixture import sakila_connection

def test_model_fetches_data(sakila_connection):
    model = TableInfoModel(sakila_connection, 'sakila', 'address')

    assert model.last_error is None
    assert isinstance(model.data, dict)

    assert len(model) == 2
    assert model['table']['TABLE_NAME'] == 'address'
    assert model['table']['TABLE_SCHEMA'] == 'sakila'
    assert len(model['create_code']) > 0

def test_model_is_empty(sakila_connection):
    model = TableInfoModel(sakila_connection, 'sakila', 'unknown model')

    assert model.last_error is None
    assert len(model) == 0

def test_model_does_not_raise_exception_on_error(sakila_connection):
    model = TableInfoModel(sakila_connection, 'sakila\' FROM x', 'address')
    assert model.last_error is not None



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid
import time

import mitzasql.ui.utils as utils
from mitzasql.history import History
from mitzasql.logger import logger
from mitzasql.db.model import (TableModel, QueryModel)
from .db_view_footer import DBViewFooter
from .query_widget import QueryWidget

class BaseDBView(urwid.Frame):
    SIGNAL_ACTION_EXIT = 'exit'
    SIGNAL_ACTION_HELP = 'help'
    SIGNAL_ACTION_RUN_QUERY = 'query_editor'
    SIGNAL_ACTION_QUERY_LOG = 'query_log'
    SIGNAL_ACTION_REFRESH = 'refresh'
    SIGNAL_ACTION_QUIT = 'quit'
    SIGNAL_MODEL_ERROR = 'model_error'
    SIGNALS = [SIGNAL_ACTION_EXIT, SIGNAL_ACTION_RUN_QUERY, SIGNAL_MODEL_ERROR]

    QUERY_HISTORY = History()

    def __init__(self, model, connection, actions=[]):
        self._model = model
        self._model_error_handler = None
        self._table = None
        self._query_editor = None
        self._connection = connection

        self._actions = []
        self._actions_to_signals = {}
        default_actions = [
               ('F1', u'F1 Help', self.SIGNAL_ACTION_HELP),
               ('F2', u'F2 Query Editor'),
               ('F5', u'F5 Refresh', self.SIGNAL_ACTION_REFRESH),
               ('F10', u'F10 Quit', self.SIGNAL_ACTION_QUIT)
               ]
        default_actions.extend(actions)
        self._prepare_actions_and_signals(default_actions)

        header = self._make_header()
        body = self._make_body()
        footer = self._make_footer()

        self._cmd_history_pending_cmd = None

        super().__init__(body, header, footer)
        signals = self.SIGNALS
        signals.extend(self._actions_to_signals.values())
        urwid.register_signal(self.__class__, signals)
        urwid.connect_signal(self, self.SIGNAL_ACTION_REFRESH,
                self.refresh_model)
        self._connect_model_signals()
        self._update_breadcrumbs()

    def _connect_model_signals(self):
        urwid.connect_signal(self._model, self._model.SIGNAL_ERROR,
                self.emit_model_error)
        urwid.connect_signal(self._model, self._model.SIGNAL_PRE_LOAD,
                self.toggle_loading_status, user_args=[True])
        urwid.connect_signal(self._model, self._model.SIGNAL_NEW_DATA,
                self.toggle_loading_status, user_args=[False])
        urwid.connect_signal(self._model, self._model.SIGNAL_LOAD,
                self.toggle_loading_status, user_args=[False])

    def _disconnect_model_signals(self):
        urwid.disconnect_signal(self._model, self._model.SIGNAL_PRE_LOAD,
                self.toggle_loading_status, user_args=[True])
        urwid.disconnect_signal(self._model, self._model.SIGNAL_NEW_DATA,
                self.toggle_loading_status, user_args=[False])
        urwid.disconnect_signal(self._model, self._model.SIGNAL_LOAD,
                self.toggle_loading_status, user_args=[False])
        urwid.disconnect_signal(self._model, self._model.SIGNAL_ERROR,
                self.emit_model_error)

    def __del__(self):
        self._disconnect_model_signals()
        urwid.disconnect_signal(self, self.SIGNAL_ACTION_REFRESH,
                self.refresh_model)
        self._disconnect_model_signals()
        self._model = None
        self._connection = None

    @property
    def model(self):
        return self._model

    def emit_model_error(self, model, error):
        urwid.emit_signal(self, self.SIGNAL_MODEL_ERROR, self, model, error)

    def set_model_error_handler(self, handler):
        if not self._model_error_handler:
            self._model_error_handler = handler
            urwid.connect_signal(self, self.SIGNAL_MODEL_ERROR, handler)

        if self._model.last_error is not None:
            self.emit_model_error(self._model, self._model.last_error)

    def toggle_loading_status(self, status, *args, **kwargs):
        self._footer.toggle_message_bar(status, message=u'Loading...')

    def refresh_model(self, emitter, action):
        self._model.reload()

    def _make_header(self):
        header = urwid.AttrMap(urwid.Text('', wrap='clip'), 'session_header')
        return header

    def _update_breadcrumbs(self):
        if self._connection.is_tcp is True:
            host = '{0}:{1}'.format(self._connection.host,
                    str(self._connection.port))
        else:
            host = self._connection.host

        session_name = self._connection.session_name

        breadcrumbs = []

        if session_name is not None:
            breadcrumbs.append(u'{0}({1})'.format(session_name, host))
        else:
            breadcrumbs.append(u'{0}'.format(host))

        if self._connection.database is not None:
            breadcrumbs.append(self._connection.database)

        if isinstance(self._model, TableModel):
            breadcrumbs.append(self._model.table_name)

        if isinstance(self._model, QueryModel):
            breadcrumbs.append(u'Query')

        breadcrumbs = '/'.join(breadcrumbs)
        utils.orig_w(self._header).set_text(u'[{0}]'.format(breadcrumbs))

    def _make_footer(self):
        command_processor = getattr(self, '_command_processor', None)
        footer = DBViewFooter(self._actions,
                command_processor=command_processor)
        urwid.connect_signal(footer, footer.SIGNAL_EXIT_COMMAND_MODE,
                self.exit_command_mode)
        return footer

    def _make_body(self):
        table = self._table_widget_cls(self._model)
        self._table = table
        self._query_editor = QueryWidget()

        pile = urwid.Pile([('weight', 70, table), ('weight', 0,
            self._query_editor)])
        return pile

    def _prepare_actions_and_signals(self, actions):
        _actions = []
        actions_to_signals = {}
        for action in actions:
            if len(action) == 3:
                key, action, signal = action
            else:
                key, action = action
                signal = None

            _actions.append((key, action))

            if signal is not None:
                actions_to_signals[key] = signal

        self._actions = sorted(_actions, key=lambda action: int(action[0][1:]))
        self._actions_to_signals = actions_to_signals

    def keypress(self, size, key):
        result = super().keypress(size, key)
        if result is None:
            return None

        for action_key, action_name in self._actions:
            if action_key.lower() == key.lower():
                if action_key in self._actions_to_signals:
                    signal = self._actions_to_signals[action_key]
                    urwid.emit_signal(self, signal, self, action_name)
                    return

        if key == 'esc':
            if self._focus_is('table'):
                urwid.emit_signal(self, self.SIGNAL_ACTION_EXIT, self)
            else:
                self._toggle_query_editor()
            return

        if hasattr(self, '_command_processor'):
            if key == 'q':
                if self._cmd_history_pending_cmd is None:
                    self._cmd_history_pending_cmd = (key, time.time())
                    return;

            if key == ':' and self._cmd_history_pending_cmd is not None:
                _none, last_key_press = self._cmd_history_pending_cmd
                now = time.time()
                self._cmd_history_pending_cmd  = None
                if (now - last_key_press) < 500:
                    if len(self._command_processor.history) > 0:
                        self._enter_command_mode(key, show_last_cmd=True)
                    return

            if self._command_processor.is_string_comand(key):
                self._enter_command_mode(key)
                return

            if self._command_processor.is_key_command(key):
                self._command_processor.execute(cmd_key=key)
                return

        # Show query editor
        if key == 'f2':
            if self._focus_is('table') and self._editor_is_visible():
                self._switch_focus('editor')
            else:
                self._toggle_query_editor()
            return

        # Run query
        if key == 'f9':
            if self._focus_is('editor') and self._editor_is_visible():
                self._run_query()
            return

        # Show previous query
        if key == 'ctrl p':
            if self._focus_is('editor') and self._editor_is_visible():
                self._show_prev_query()
            return

        # Show next query
        if key == 'ctrl n':
            if self._focus_is('editor') and self._editor_is_visible():
                self._show_next_query()
            return

        # Clear editor
        if key == 'ctrl f9':
            if self._focus_is('editor') and self._editor_is_visible():
                self._query_editor.clear()
            return
        return key

    def _editor_is_visible(self):
        widget, size = self._body.contents[1]
        return size[1] != 0

    def _run_query(self):
        query = self._query_editor.query
        self.QUERY_HISTORY.append(query)
        urwid.emit_signal(self, self.SIGNAL_ACTION_RUN_QUERY, self, query)

    def _show_prev_query(self):
        query = self.QUERY_HISTORY.prev
        if not query:
            return
        self._query_editor.query = query

    def _show_next_query(self):
        query = self.QUERY_HISTORY.next
        if not query:
            return
        self._query_editor.query = query

    def _toggle_query_editor(self):
        widget, size = self._body.contents[1]
        if size[1] == 0:
            new_size = ('weight', 20)
            focus_pos = 1
        else:
            new_size = ('weight', 0)
            focus_pos = 0
        self._body.contents[1] = (widget, new_size)
        self._body.focus_position = focus_pos

    def _focus_is(self, widget_name):
        focus_pos = self._body.focus_position
        if focus_pos == 0:
            return widget_name == 'table'

        return widget_name == 'editor'

    def _switch_focus(self, widget_name):
        if widget_name == 'table':
            self._body.focus_position = 0
            return
        self._body.focus_position = 1

    def _enter_command_mode(self, key, show_last_cmd=False):
        self.focus_position = 'footer'
        self._footer.toggle_command_mode(True, key, show_last_cmd=show_last_cmd)

    def exit_command_mode(self, *args, **kwargs):
        self.focus_position = 'body'



import pytest

from mitzasql.history import History

def test_history_returns_last_item():
    history = History()
    history.append('item1')
    history.append('item2')
    history.append('item3')

    last = history.last
    assert last == 'item3'

def test_history_returns_prev_and_next_items():
    history = History()
    history.append('item1')
    history.append('item2')
    history.append('item3')

    prev = history.prev
    assert prev == 'item3'
    prev = history.prev
    assert prev == 'item2'
    prev = history.prev
    assert prev == 'item1'

    next_ = history.next
    assert next_ == 'item2'
    next_ = history.next
    assert next_ == 'item3'



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from .list import List
from .action_bar_pile_container import ActionBarPileContainer
from .action_bar import ActionBar

class SavedSessionsList(ActionBarPileContainer):
    # Signals handled by the select session UI screen
    SIGNAL_CONNECT = 'connect'
    SIGNAL_CREATE_NEW = 'new'
    SIGNAL_EDIT = 'edit'
    SIGNAL_DELETE = 'delete'
    SIGNAL_QUIT = 'quit'

    def __init__(self, sessions_list):
        list = List(sessions_list)
        self._list = list
        actions = [
                ('N', u'New', self.SIGNAL_CREATE_NEW),
                ('C', u'Connect', self.SIGNAL_CONNECT),
                ('E', u'Edit', self.SIGNAL_EDIT),
                ('D', u'Delete', self.SIGNAL_DELETE),
                ('Q', u'Quit', self.SIGNAL_QUIT),
                ]
        super().__init__(urwid.BoxAdapter(list, 10), actions, 10,
                line_box=True, line_box_title=u'Select Session')
        urwid.connect_signal(list, List.SIGNAL_SELECTED, self._emit_connect_signal)

    def _emit_connect_signal(self, emitter, *args, **kwargs):
        urwid.emit_signal(self, self.SIGNAL_CONNECT, self)

    def refresh(self, data):
        self._list.reset(data)

    @property
    def session(self):
        return self._list.focus.item_value





# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from .query_editor import QueryEditor

class QueryWidget(urwid.AttrMap):
    def __init__(self):
        editor = QueryEditor()
        container = urwid.Filler(urwid.AttrMap(editor, ''), valign='top')
        title = u'[F9: Run. Ctrl-F9: Clear. Esc: Close. Ctrl-P/Ctrl-N: Query history]'
        line_box = urwid.LineBox(container, title=title,
                title_align='right')

        super().__init__(line_box, 'linebox')

    @property
    def query(self):
        return self.base_widget.edit_text

    @query.setter
    def query(self, query):
        self.base_widget.edit_text = query

    def clear(self):
        self.base_widget.edit_text = ''

    def keypress(self, size, key):
        key = super().keypress(size, key)
        # Disable the up key. It interferes with the parent pile container
        if key == 'up':
            return None
        return key



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from .table import Table

from mitzasql.db.model import TableModel as MysqlTableModel

class MysqlTable(Table):
    def __init__(self, model):
        if not isinstance(model, MysqlTableModel):
            raise TypeError("Wrong model type")
        super().__init__(model)

        urwid.connect_signal(model, model.SIGNAL_NEW_DATA, self.render_more)

    def __del__(self):
        urwid.disconnect_signal(self._model, self._model.SIGNAL_NEW_DATA, self.render_more)
        super().__del__()

    def render_more(self, model, data, data_length):
        for row in data:
            self._body.add_row(row)

    def _scroll_rows(self):
        if self._model.loaded_rowcount < len(self._model):
            remaining_rows_to_scroll = (self._model.loaded_rowcount - 1) - self._focused_row_index

            if remaining_rows_to_scroll > 0 and remaining_rows_to_scroll < 50:
                self._model.load_next_set()
            elif remaining_rows_to_scroll < 0:
                self._model.load_more(self._focused_row_index)

        if self._focused_row_index <= (self._model.loaded_rowcount - 1):
            return super()._scroll_rows()




import pytest
import urwid

from mitzasql.ui.widgets.db_view import DBView
from mitzasql.db.model import DatabasesModel
from ..db.connection_fixture import connection

def test_keypresses(connection):
    model = DatabasesModel(connection)
    view = DBView(model, connection)

    size = (80, 30)

    emitter = None
    database = None
    def test_db_is_selected(emitter_, database_):
        nonlocal emitter
        nonlocal database
        emitter = emitter_
        database = database_

    urwid.connect_signal(view, view.SIGNAL_ACTION_SELECT_DB,
            test_db_is_selected)

    view.keypress(size, 'down')
    view.keypress(size, 'down')
    view.keypress(size, 'down')
    view.keypress(size, 'up')
    view.keypress(size, 'up')
    view.keypress(size, 'up')

    view.keypress(size, 'enter')

    assert isinstance(emitter, DBView)
    assert database is not None
    old_database = database

    view.keypress(size, 'down')
    view.keypress(size, 'enter')
    assert database is not None
    assert database != old_database


def test_db_view_searches_for_database(connection):
    model = DatabasesModel(connection)
    view = DBView(model, connection)

    size = (80, 30)

    result = view.search_db('sakila')
    assert result is not None

    index, row = result
    assert index > 0
    assert 'sakila' in row

    result = view.search_db('non existent database')
    assert result is None

def test_db_view_refreshes_table(connection):
    model = DatabasesModel(connection)
    view = DBView(model, connection)
    size = (80, 30)

    initial_focus = view._table._focused_row_index

    view.keypress(size, 'down')
    view.keypress(size, 'down')

    new_focus = view._table._focused_row_index

    assert new_focus > initial_focus

    model.reload()
    new_focus = view._table._focused_row_index
    assert new_focus == initial_focus



import pytest

from mitzasql.db.model import ViewInfoModel
from .connection_fixture import sakila_connection

def test_model_fetches_data(sakila_connection):
    model = ViewInfoModel(sakila_connection, 'sakila', 'customer_list')

    assert model.last_error is None
    assert isinstance(model.data, dict)

    assert len(model) > 0
    assert model['TABLE_SCHEMA'] == 'sakila'
    assert model['TABLE_NAME'] == 'customer_list'
    assert len(model['VIEW_DEFINITION']) > 0

def test_model_is_empty(sakila_connection):
    model = ViewInfoModel(sakila_connection, 'sakila', 'unknown view')

    assert model.last_error is None
    assert len(model) == 0

def test_model_does_not_raise_exception_on_error(sakila_connection):
    model = ViewInfoModel(sakila_connection, 'sakila\' FROM x', 'customer_create_date')

    assert model.last_error is not None



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.ui.widgets.create_new_session_dialog import CreateNewSessionDialog
from mitzasql.ui.widgets.saved_sessions_list import SavedSessionsList
from mitzasql.ui.widgets.edit_session_form import EditSessionForm
from mitzasql.ui.screens.base_widgets_factory import BaseWidgetsFactory
from . import states

class WidgetsFactory(BaseWidgetsFactory):
    """Creates and caches widgets for the session select screen"""

    def _get_factory_methods(self):
        factory_methods = {}
        factory_methods['create_session_dialog'] = self.create_create_session_dialog
        factory_methods['sessions_list'] = self.create_sessions_list
        factory_methods['create_session_form'] = self.create_create_session_form

        return factory_methods

    def create_create_session_dialog(self):
        dialog = CreateNewSessionDialog()
        self._connect_signal(dialog, CreateNewSessionDialog.SIGNAL_YES,
                self.change_fsm_state, user_args=['yes'])
        self._connect_signal(dialog, CreateNewSessionDialog.SIGNAL_QUIT,
                self.change_fsm_state, user_args=['quit'])
        return dialog

    def create_sessions_list(self, sessions):
        list = SavedSessionsList(sessions)
        self._connect_signal(list, SavedSessionsList.SIGNAL_QUIT,
                self.change_fsm_state, user_args=['quit'])
        self._connect_signal(list, SavedSessionsList.SIGNAL_CREATE_NEW,
                self.change_fsm_state, user_args=['create'])
        self._connect_signal(list, SavedSessionsList.SIGNAL_CONNECT,
                self.change_fsm_state, user_args=['connect'])
        self._connect_signal(list, SavedSessionsList.SIGNAL_EDIT,
                self.change_fsm_state, user_args=['edit'])
        self._connect_signal(list, SavedSessionsList.SIGNAL_DELETE,
                self.change_fsm_state, user_args=['delete'])
        return list

    def create_create_session_form(self, form_data):
        form = EditSessionForm(form_data)
        self._connect_signal(form, form.SIGNAL_SAVE, self.change_fsm_state,
                user_args=['save'])
        self._connect_signal(form, form.SIGNAL_TEST, self.change_fsm_state,
                user_args=['test'])
        self._connect_signal(form, form.SIGNAL_CONNECT, self.change_fsm_state,
                user_args=['connect'])
        self._connect_signal(form, form.SIGNAL_CANCEL, self.change_fsm_state,
                user_args=['cancel'])
        return form



import pytest

from mitzasql.table_filter_parser import Parser

def test_parser():
    p = Parser("1 2")
    tokens = p.parse()
    assert tokens[0] == '1'
    assert tokens[1] == '2'

    p = Parser("abc defg")
    tokens = p.parse()
    assert tokens[0] == 'abc'
    assert tokens[1] == 'defg'

    p = Parser("abc       defg")
    tokens = p.parse()
    assert tokens[0] == 'abc'
    assert tokens[1] == 'defg'

    p = Parser("  abc       defg")
    tokens = p.parse()
    assert tokens[0] == 'abc'
    assert tokens[1] == 'defg'

    p = Parser("  abc       defg  ")
    tokens = p.parse()
    assert tokens[0] == 'abc'
    assert tokens[1] == 'defg'

    p = Parser("abc       defg  ")
    tokens = p.parse()
    assert tokens[0] == 'abc'
    assert tokens[1] == 'defg'

    p = Parser("'abc   defg' 123")
    tokens = p.parse()
    assert tokens[0] == 'abc   defg'
    assert tokens[1] == '123'

    p = Parser("'abc   defg''123'")
    tokens = p.parse()
    assert tokens[0] == 'abc   defg123'

    p = Parser("'ab\\'c   defg' 123")
    tokens = p.parse()
    assert tokens[0] == 'ab\'c   defg'
    assert tokens[1] == '123'

    p = Parser('"ab\\"c   de\\"fg" 123')
    tokens = p.parse()
    assert tokens[0] == 'ab"c   de"fg'
    assert tokens[1] == '123'

    p = Parser('\\r')
    tokens = p.parse()
    assert tokens[0] == '\\r'

    p = Parser('\\r 123 \\n\\r\\tabc')
    tokens = p.parse()
    assert tokens[0] == '\\r'
    assert tokens[1] == '123'
    assert tokens[2] == '\\n\\r\\tabc'

    p = Parser("'2017-09-02 12:12:12' '2018-02-02 12:13:13'")
    tokens = p.parse()
    assert tokens[0] == '2017-09-02 12:12:12'
    assert tokens[1] == '2018-02-02 12:13:13'



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
from collections import OrderedDict

import mysql.connector.errors as errors
from mysql.connector import (FieldType, FieldFlag)

# Column types
int_ctypes = ['int', 'tinyint', 'smallint', 'mediumint', 'bigint', 'bit']
real_ctypes = ['float', 'double', 'decimal']
text_ctypes = ['char', 'varchar', 'tinytext', 'text', 'mediumtext', 'longtext',
        'json', 'enum', 'set']
binary_ctypes = ['binary', 'varbinary', 'tinyblob', 'blob', 'mediumblob',
        'longblob']
temporal_ctypes = ['date', 'time', 'year', 'datetime', 'timestamp']
spatial_ctypes = ['point', 'linestring', 'polygon', 'geometry', 'multipoint',
        'multilinestring', 'multipolygon', 'geometrycollection']

# Field types
int_ftypes = [FieldType.TINY, FieldType.SHORT, FieldType.LONG,
        FieldType.LONGLONG, FieldType.INT24, FieldType.BIT]
real_ftypes = [FieldType.DECIMAL, FieldType.NEWDECIMAL, FieldType.FLOAT, FieldType.DOUBLE]
text_ftypes = [FieldType.VARCHAR, FieldType.VAR_STRING, FieldType.ENUM,
        FieldType.SET, FieldType.STRING, FieldType.JSON]
binary_ftypes = [FieldType.TINY_BLOB, FieldType.MEDIUM_BLOB,
        FieldType.LONG_BLOB, FieldType.BLOB]
temporal_ftypes = [FieldType.TIMESTAMP, FieldType.TIME, FieldType.DATE,
        FieldType.YEAR, FieldType.DATETIME, FieldType.NEWDATE]
spatial_ftypes = [FieldType.GEOMETRY]

# The required number of chars to display a specific type
field_types_length = {
        FieldType.DECIMAL: 66,
        FieldType.TINY: 4,
        FieldType.SHORT: 6,
        FieldType.LONG: 11,
        FieldType.FLOAT: 12,
        FieldType.DOUBLE: 24,
        FieldType.NULL: 4,
        FieldType.TIMESTAMP: 19,
        FieldType.LONGLONG: 24,
        FieldType.INT24: 8,
        FieldType.DATE: 10,
        FieldType.TIME: 10,
        FieldType.DATETIME: 19,
        FieldType.YEAR: 4,
        FieldType.NEWDATE: 19,
        FieldType.VARCHAR: sys.maxsize,
        FieldType.BIT: sys.maxsize,
        FieldType.JSON: sys.maxsize,
        FieldType.NEWDECIMAL: 66,
        FieldType.ENUM: 32,
        FieldType.SET: 32,
        FieldType.TINY_BLOB: sys.maxsize,
        FieldType.MEDIUM_BLOB: sys.maxsize,
        FieldType.LONG_BLOB: sys.maxsize,
        FieldType.BLOB: sys.maxsize,
        FieldType.VAR_STRING: sys.maxsize,
        FieldType.STRING: sys.maxsize,
        FieldType.GEOMETRY: sys.maxsize
        }

# Types for which to detect their string representation length
auto_detect_field_type_length = [
        FieldType.LONGLONG,
        FieldType.DECIMAL,
        FieldType.DOUBLE,
        FieldType.VARCHAR,
        FieldType.BIT,
        FieldType.NEWDECIMAL,
        FieldType.ENUM,
        FieldType.SET,
        FieldType.VAR_STRING,
        FieldType.STRING,
        FieldType.GEOMETRY
        ]

class QuerySchema:
    def __init__(self, cursor, data_sample):
        self.cursor = cursor
        self.data = data_sample
        self.schema = {}

    @property
    def columns(self):
        '''Returns a dictionary containing the column names as keys and
        metadata as their value:

        columns['name'] = {
            'friendly_type': 'VAR_STRING',
            'driver_type': 127,
            'flags': { # dictionary of mysql flags }
        }

        '''
        columns = []

        column_index = 0
        for column_data in self.cursor.description:
            name = column_data[0]
            driver_type = column_data[1]
            flags = get_column_flags(column_data[7])

            column = {
                    'name': name,
                    'flags': flags
                    }

            try:
                c_type = self.schema[name]['type']
                column.update({
                        'is_int': c_type in int_ctypes,
                        'is_real': c_type in real_ctypes,
                        'is_text': c_type in text_ctypes,
                        'is_binary': c_type in binary_ctypes,
                        'is_temporal': c_type in temporal_ctypes,
                        'is_spatial': c_type in spatial_ctypes
                        })
            except KeyError as e:
                column.update({
                        'is_int': driver_type in int_ftypes,
                        'is_real': driver_type in real_ftypes,
                        'is_text': driver_type in text_ftypes,
                        'is_binary': driver_type in binary_ftypes,
                        'is_temporal': driver_type in temporal_ftypes,
                        'is_spatial': driver_type in spatial_ftypes
                        })

            try:
                max_len = self.schema[name]['max_len']
            except KeyError as e:
                max_len = None

            if max_len is None:
                if driver_type in auto_detect_field_type_length:
                    max_len = auto_detect_column_length(column_index, self.data)

                if max_len is None:
                    max_len = field_types_length[driver_type]

            if len(name) >= max_len:
                max_len = len(name)
            elif max_len > 66:
                max_len = 66

            # The max_len is used by the table widget as width for the the
            # columns. Even though this is related to the view logic, we do it
            # here to optimize the rendering
            column['max_len'] = max_len + 2
            columns.append(column)
            column_index += 1
        return columns

class TableSchema(QuerySchema):
    def __init__(self, connection, table_name):
        self._con = connection
        self._table_name = table_name
        self.schema = self._table_schema()
        self.cursor = None
        self.data = None

    def __iter__(self):
        for row in self.schema.items():
            yield row

    def column_is_text(self, type_):
        return type_ in text_ctypes

    def column_is_spatial(self, type_):
        return type_ in spatial_ctypes

    def _table_schema(self):
        query = '''
        SELECT
            column_name,
            column_default,
            is_nullable,
            data_type,
            character_maximum_length,
            numeric_precision,
            numeric_scale,
            character_set_name,
            collation_name,
            column_key,
            extra
        FROM `information_schema`.`columns`
        WHERE
            table_schema = %(db_name)s
            AND
            table_name = %(table_name)s
        ORDER BY ORDINAL_POSITION ASC
        '''
        cursor = self._con.query(query, {
            'db_name': self._con.database,
            'table_name': self._table_name
            })
        data = cursor.fetchall()
        schema = OrderedDict()
        for row in data:
            column_name, default, nullable, _type, char_max_len, num_precision, \
                    num_scale, charset, collation, key, extra = row

            if nullable == 'YES':
                nullable = True
            else:
                nullable = False

            schema[column_name] = {
                    'type': _type,
                    'charset': charset,
                    'collation': collation,
                    'default': default,
                    'nullable': nullable,
                    'key': key,
                    'extra': extra
                    }

            c_len = None
            if _type in int_ctypes or _type in real_ctypes:
                if num_precision is not None or num_scale is not None:
                    # Add 1 character to len to account for any coma that might exist
                    c_len = int(num_precision or 0) + int(num_scale or 0) + 1
            elif _type in text_ctypes or _type in binary_ctypes:
                if char_max_len is not None:
                    c_len = int(char_max_len)

            schema[column_name]['max_len'] = c_len
        return schema


def auto_detect_column_length(c_index, sample):
    '''Get the max length for the item in column `c_index` in the first
    `MAX_ROW_FOR_COLUMN_LENGTH_DETECTION`
    '''
    if len(sample) == 0:
        return None
    column = [str(r[c_index]) for r in sample]
    item = max(column, key=len)
    return len(item)


def get_column_flags(flag):
    flags = {}
    for name in FieldFlag.desc:
        (value, description) = FieldFlag.desc[name]
        if flag & value:
            flags[name] = {
                    'name': name,
                    'value': value,
                    'desc': description
                    }
    return flags



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.state_machine import StateMachine
from mitzasql.ui.screens.screen import Screen
from mitzasql.ui.widgets.saved_sessions_list import SavedSessionsList
from mitzasql.ui.widgets.edit_session_form import EditSessionForm
from .widgets_factory import WidgetsFactory
from . import states

class SessionSelect(Screen):
    """The session select UI screen

    Implements the logic for creating, editing, deleting and selecting the
    session to connect to. When the user selects a session to connect to, the
    screen calls the main application instance with the session name or
    connection details.
    Child widgets are created using a widget factory and their state is managed
    with the help of a FSM"""

    # Signals handled by the main application instance
    SIGNAL_TEST_CONNECTION = 'test_connection'
    SIGNAL_CONNECT = 'connect'

    def __init__(self, sessions_registry):
        super().__init__([self.SIGNAL_CONNECT, self.SIGNAL_TEST_CONNECTION])
        self._sessions_registry = sessions_registry
        self._state_machine = self._init_state_machine()
        self._widgets_factory = WidgetsFactory(self._state_machine)
        self._state_machine.set_initial_state(states.STATE_INITIAL)
        self._state_machine.run()

    def __del__(self):
        self.focused_widget = None
        self.view = None
        self._widgets_factory = None

    def _init_state_machine(self):
        state_machine = StateMachine('session_select')
        state_machine.add_state(states.STATE_INITIAL, self.initial_state, {
            'has_sessions': states.STATE_SHOW_SESSIONS_LIST,
            'no_sessions': states.STATE_SHOW_CREATE_SESSION_DIALOG
            })
        state_machine.add_state(states.STATE_SHOW_CREATE_SESSION_DIALOG,
                self.show_create_new_session_dialog, {
                    'yes': states.STATE_SHOW_EDIT_SESSION_FORM,
                    'quit': states.STATE_QUIT
                    })
        state_machine.add_state(states.STATE_SHOW_SESSIONS_LIST,
                self.show_sessions_list, {
                    'quit': states.STATE_QUIT,
                    'create': states.STATE_SHOW_EDIT_SESSION_FORM,
                    'edit': states.STATE_SHOW_EDIT_SESSION_FORM,
                    'delete': states.STATE_DELETE_SESSION,
                    'connect': states.STATE_CONNECT
                    })
        state_machine.add_state(states.STATE_QUIT, self.quit)
        state_machine.add_state(states.STATE_CANCEL_EDIT_SESSION,
                self.cancel_edit_session, {
                    'quit': states.STATE_QUIT,
                    'show_sessions': states.STATE_SHOW_SESSIONS_LIST
                    })
        state_machine.add_state(states.STATE_SHOW_EDIT_SESSION_FORM,
                self.show_edit_session_form, {
                    'save': states.STATE_SAVE_SESSION,
                    'test': states.STATE_TEST_CONNECTION,
                    'cancel': states.STATE_CANCEL_EDIT_SESSION,
                    'connect': states.STATE_CONNECT
                    })
        state_machine.add_state(states.STATE_SAVE_SESSION,
                self.save_connection, {
                    'invalid': states.STATE_SHOW_EDIT_SESSION_FORM,
                    'valid': states.STATE_SHOW_SESSIONS_LIST
                    })
        state_machine.add_state(states.STATE_DELETE_SESSION,
                self.delete_session, {
                    'deleted': states.STATE_INITIAL
                    })
        state_machine.add_state(states.STATE_TEST_CONNECTION,
                self.test_connection, {
                    'invalid': states.STATE_SHOW_EDIT_SESSION_FORM,
                    'error': states.STATE_SHOW_EDIT_SESSION_FORM,
                    'success': states.STATE_SHOW_EDIT_SESSION_FORM
                    })
        state_machine.add_state(states.STATE_CONNECT, self.connect, {
                    'error_goto_form': states.STATE_SHOW_EDIT_SESSION_FORM,
                    'error_goto_list': states.STATE_SHOW_SESSIONS_LIST
                    })

        return state_machine

    def initial_state(self):
        if self._sessions_registry.is_empty():
            self._state_machine.change_state('no_sessions')
            return

        self._state_machine.change_state('has_sessions')

    def cancel_edit_session(self, *args, **kwargs):
        if self._sessions_registry.is_empty():
            self._state_machine.change_state('quit')
            return
        self._state_machine.change_state('show_sessions')

    def show_create_new_session_dialog(self, *args, **kwargs):
        self.focused_widget = self._widgets_factory.create('create_session_dialog')
        self.view.original_widget = urwid.Filler(urwid.Padding(self.focused_widget, align='center',
            width=('relative', 50)))

    def show_connection_error(self, *args, **kwargs):
        self.focused_widget = self._widgets_factory.create('dialog', u'Error', title=u'Connection error')
        self.view.original_widget = urwid.Filler(urwid.Padding(self.focused_widget, align='center', width=('relative', 50)))

    def show_sessions_list(self, *args, list_has_message=False):
        sessions = self._sessions_registry.sessions
        self.focused_widget = self._widgets_factory.create('sessions_list', sessions)
        self.view.original_widget = urwid.Filler(self.focused_widget)
        if list_has_message is False:
            self.focused_widget.clear_status_message()

    def delete_session(self, emitter, action):
        session_name = emitter.session
        del self._sessions_registry[session_name]
        self._sessions_registry.save()
        self._state_machine.change_state('deleted')

    def show_edit_session_form(self, *args, form_has_message=False, form_data={}):
        new_session = True
        if len(args) == 2:
            emitter, action = args
            if isinstance(emitter, SavedSessionsList) and action == 'Edit':
                new_session = False
                session_name = emitter.session
                form_data = self._sessions_registry[session_name]
                form_data['name'] = session_name
            elif isinstance(emitter, SavedSessionsList) and action == 'New':
                if form_has_message is False:
                    form_data = {}

        form = self._widgets_factory.create('create_session_form', form_data)

        if new_session is True:
            form_title = 'Create session'
        else:
            form_title = 'Edit session'

        form.editing_new_session = new_session
        form.title = form_title

        self.focused_widget = form
        self.view.original_widget = urwid.Filler(self.focused_widget)

        if form_has_message is False:
            form.clear_status_message()
            form.reset_focus()

    def test_connection(self, *args, **kwargs):
        form, *args = args
        form_data = form.data

        validation_error = self.validate_connection_data(form_data)
        if validation_error is not None:
            form.set_status_message(validation_error, error=True)
            self._state_machine.change_state('invalid',
                    form_has_message=True, form_data=form_data)
            return

        form.clear_status_message()
        urwid.emit_signal(self, self.SIGNAL_TEST_CONNECTION, self, form_data)

    def connect(self, *args, **kwargs):
        widget, *args = args

        if isinstance(widget, SavedSessionsList):
            session_name = widget.session
            urwid.emit_signal(self, self.SIGNAL_CONNECT, self, session_name)
            return

        if isinstance(widget, EditSessionForm):
            data = widget.data
            validation_error = self.validate_connection_data(data)
            if validation_error is not None:
                widget.set_status_message(validation_error, error=True)
                self._state_machine.change_state('error_goto_form',
                        form_has_message=True, form_data=data)
                return
            urwid.emit_signal(self, self.SIGNAL_CONNECT, self, None, data)
            return

        raise RuntimeError('The current widget should not trigger a connect signal: {0}'.format(str(widget)))


    def set_connection_result(self, result, error=None, session_name=None, connection_data=None):
        if result is False:
            self.focused_widget.set_status_message(error, error=True)
            if session_name is None:
                self._state_machine.change_state('error_goto_form',
                        form_has_message=True, form_data=connection_data)
            else:
                self._state_machine.change_state('error_goto_list',
                        list_has_message=True)
            return
        # nothing to do if result is True, our job is done

    def set_progress(self, message):
        widget = self.focused_widget
        widget.set_status_message(message)

    def set_connection_test_result(self, result, error=None):
        form = self.focused_widget
        form_data = form.data
        if result is False:
            form.set_status_message(error, error=True)
            self._state_machine.change_state('error', form_has_message=True,
                    form_data=form_data)
            return
        form.set_status_message(u'Connection was successful')
        self._state_machine.change_state('success', form_data=form_data,
                form_has_message=True)

    def save_connection(self, *args, **kwargs):
        form, *args = args
        form_data = form.data

        validation_error = self.validate_connection_data(form_data)
        if validation_error is not None:
            form.set_status_message(validation_error, error=True)
            self._state_machine.change_state('invalid',
                    form_has_message=True, form_data=form_data)
            return

        if 'old_name' in form_data and form_data['old_name'] is not None:
            # Handle update session name
            old_name = form_data['old_name']
            del form_data['old_name']
            self._sessions_registry.edit(old_name, form_data)
        else:
            if 'old_name' in form_data and form_data['old_name'] is None:
                del form_data['old_name']
            self._sessions_registry.add(form_data)
        self._sessions_registry.save()
        self._state_machine.change_state('valid')

    def validate_connection_data(self, data):
        if len(data['name']) == 0:
            return u'Session name is required'

        if len(data['host']) == 0:
            return u'Host is required'

        host = data['host']
        if not host.startswith('tcp://') and not host.startswith('unix://'):
            return u'Host is missing protocol (tcp://, unix://)'

        host_is_tcp = False

        if host.startswith('tcp://'):
            host_is_tcp = True
            host = host[6:]
        elif host.startswith('unix://'):
            host = host[7:]

        if len(host) == 0:
            return u'Host must not be empty'

        if host_is_tcp is True and ('port' not in data or len(data['port']) == 0):
            return u'Port is required'

        if 'username' not in data or len(data['username']) == 0:
            return u'Username is required'

        return None



import pytest

from mitzasql.db.model import TriggerModel
from .connection_fixture import sakila_connection

def test_model_fetches_data(sakila_connection):
    model = TriggerModel(sakila_connection, 'sakila', 'customer_create_date')

    assert model.last_error is None
    assert isinstance(model.data, dict)

    assert model['TRIGGER_SCHEMA'] == 'sakila'
    assert model['TRIGGER_NAME'] == 'customer_create_date'
    assert model['EVENT_MANIPULATION'] == 'INSERT'
    assert model['EVENT_OBJECT_TABLE'] == 'customer'
    assert len(model['ACTION_STATEMENT']) > 0

def test_model_is_empty(sakila_connection):
    model = TriggerModel(sakila_connection, 'sakila', 'unknown trigger')

    assert model.last_error is None
    assert len(model) == 0

def test_model_does_not_raise_exception_on_error(sakila_connection):
    model = TriggerModel(sakila_connection, 'sakila\' FROM x', 'customer_create_date')

    assert model.last_error is not None



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from mitzasql.history import History

class CommandError(Exception):
    pass

class Autocomplete:
    def __init__(self, suggestions, word):
        self._suggestions = suggestions
        self._suggestions.sort()
        self._index = None
        self._keyword = word
        self._matches = self._make_matches()
        self._last_dir = None

    def _make_matches(self):
        matches = []
        for sug in self._suggestions:
            if sug.lower().startswith(self._keyword.lower()):
                matches.append(sug)
        return matches

    def suggestion(self, direction):
        if self._last_dir is None:
            self._last_dir = direction
        else:
            if self._last_dir != direction:
                if direction == 'forward':
                    self._index += 2
                else:
                    self._index -= 2
            self._last_dir = direction

        if direction == 'forward':
            return self.next_suggestion()
        return self.prev_suggestion()

    def next_suggestion(self):
        if not len(self._matches):
            return None
        if self._index is None or self._index > len(self._matches) - 1:
            self._index = 0
        match = self._matches[self._index]
        self._index += 1
        return match

    def prev_suggestion(self):
        if not len(self._matches):
            return None
        if self._index is None or self._index < 0:
            self._index = len(self._matches) - 1

        match = self._matches[self._index]
        self._index -= 1
        return match

class BaseCmdProcessor:
    SIGNAL_ERROR = 'error'
    '''Support basic VIM style commands'''
    def __init__(self):
        self.history = History()
        self._cmd_strings = {}
        self.cmd_args_suggestions = {}
        self._colon_cmds = [
                ('q', 'quit', self._quit)
                ]
        self._cmd_keys = {}
        self._autocomplete = None
        urwid.register_signal(self.__class__, self.SIGNAL_ERROR)

    def is_string_comand(self, key):
        return key in self._cmd_strings

    def is_key_command(self, key):
        return key in self._cmd_keys

    def colon_handler(self, cmd):
        self.history.append(cmd)
        if not cmd or len(cmd) == 0:
            return

        cmd, *args = cmd.split(' ')
        for colon_cmd in self._colon_cmds:
            if cmd not in colon_cmd[0:-1]:
                continue
            handler = colon_cmd[-1]
            return handler(*args)

        self._emit_error(u'Command not found!')

    def execute(self, cmd_key=None, str_cmd_marker=None, cmd=None):
        if cmd_key in self._cmd_keys:
            handler = self._cmd_keys[cmd_key]
            return handler()

        if str_cmd_marker in self._cmd_strings:
            handler = self._cmd_strings[str_cmd_marker]
            return handler(cmd)

    def _quit(self):
        raise urwid.ExitMainLoop()

    def reset_autocomplete(self):
        self._autocomplete = None

    def _emit_error(self, error):
        urwid.emit_signal(self, self.SIGNAL_ERROR, self, error)

    def autocomplete(self, word, command=False, argument=False,
            command_name=None, direction='forward'):

        if self._autocomplete is None:
            if command is True:
                suggestions = [cmd[1] for cmd in self._colon_cmds]
                self._autocomplete = Autocomplete(suggestions, word)

            if argument is True:
                try:
                    suggestions = self.cmd_args_suggestions[command_name]
                except KeyError:
                    suggestions = []
                self._autocomplete = Autocomplete(suggestions, word)

            return self._autocomplete.suggestion(direction)

        return self._autocomplete.suggestion(direction)

class SearchCmdProcessor:
    def __init__(self, search_callback):
        self._last_keyword = None
        self._last_search_result = None
        self._max_search_pos = None
        self._search_callback = search_callback

    def search(self, keyword):
        if keyword is None or len(keyword) == 0:
            return
        self._last_keyword = keyword
        self._last_search_result = self._search_callback(keyword)
        if self._last_search_result is not None:
            self._max_search_pos = self._last_search_result[0]
            return
        self._emit_error(u'Nothing found')

    def search_next(self):
        if self._last_keyword is None:
            return
        if self._last_search_result is None:
            return
        index, row = self._last_search_result
        result = self._search_callback(self._last_keyword,
                pos=index + 1)
        if result is not None:
            self._last_search_result = result
            if self._last_search_result[0] > self._max_search_pos:
                self._max_search_pos = self._last_search_result[0]
        else:
            self._last_search_result = (0, self._last_search_result[1])
            return self.search_next()

    def search_prev(self):
        if self._last_keyword is None:
            return
        if self._last_search_result is None:
            return
        index, row = self._last_search_result
        result = self._search_callback(self._last_keyword,
                pos=index, reverse=True)
        if result is not None:
            self._last_search_result = result
            if self._last_search_result[0] > self._max_search_pos:
                self._max_search_pos = self._last_search_result[0]
        else:
            self._last_search_result = (self._max_search_pos + 1, self._last_search_result[1])
            return self.search_prev()




# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import time

import urwid
import mysql.connector
from mysql.connector import (errorcode, errors)

from mitzasql.logger import (LoggerMixin, logger)

class Connection(LoggerMixin):
    SIGNAL_EXCEPTION = 'exception'
    def __init__(self, connection_data, session_name=None):
        self.set_log_prefix('Connection')
        self.is_tcp = False
        self.database = None
        self._retry_count = 0
        self.query_log = []
        self.con = self._create_mysql_connection(connection_data)
        self._connection_data = connection_data
        try:
            self.session_name = connection_data['name']
        except KeyError:
            self.session_name = session_name
        urwid.register_signal(self.__class__, [self.SIGNAL_EXCEPTION])

    def _create_mysql_connection(self, connection_data):
        kwargs = {
                'user': connection_data['username'],
                'password': connection_data['password'],
                'connection_timeout': 5,
                'autocommit': True,
                'use_pure': True
                }

        protocol, host = self._extract_host(connection_data['host'])
        if protocol == 'tcp':
            kwargs['host'] = host
            kwargs['port'] = connection_data['port']
            self.is_tcp = True
        else:
            kwargs['unix_socket'] = host

        if 'database' in connection_data and len(connection_data['database']) > 0:
            kwargs['database'] = connection_data['database']
            self.database = connection_data['database']

        kwargs['get_warnings'] = True

        con = mysql.connector.connect(**kwargs)
        self._retry_count = 0
        return con

    @property
    def fresh(self):
        con_data = self._connection_data
        if self.database:
            con_data['database'] = self.database
        con = self.__class__(con_data, self.session_name)
        return con

    def _retry_connection(self):
        self.log_info("Disconnected. Retrying mysql connection...")
        if self._retry_count >= 5:
            return False

        try:
            self.con = self._create_mysql_connection(self._connection_data)
            if self.database:
                self.change_db(self.database)
            return True
        except errors.OperationalError as e:
            self._retry_count += 1
            time.sleep(1)
            return self._retry_connection()

    def _extract_host(self, host):
        if host.startswith('unix://'):
            protocol = 'unix'
            hostname = host[7:]
        else:
            protocol = 'tcp'
            hostname = host[6:]

        return (protocol, hostname)

    @property
    def host(self):
        protocol, host = self._extract_host(self._connection_data['host'])
        return host

    @property
    def port(self):
        return self._connection_data['port']

    def close(self):
        return self.con.close()

    @property
    def cursor(self):
        return self.con.cursor(buffered=True)

    @property
    def dict_cursor(self):
        return self.con.cursor(buffered=True, dictionary=True)

    def query(self, query, params=None, dictionary=False):
        self.query_log.append((query, params))
        self.log_debug('Query: %s. Params: %s', str(query), str(params))
        try:
            if dictionary is False:
                cursor = self.cursor
            else:
                cursor = self.dict_cursor
            cursor.execute(query, params=params)
            return cursor
        except errors.OperationalError as e:
            # Retry connection if exception is "MySQL Connection not available"
            self.log_exception('Query exception: %s', e)
            if self._retry_connection():
                return self.query(query, params=params, dictionary=dictionary)
            raise e
        except errors.Error as e:
            self.log_exception('Query exception: %s', e)
            raise e

    def change_db(self, name):
        query = 'USE {0}'.format(name);
        self.query(query);
        self.database = name

    def is_fatal_error(self, error):
        if isinstance(error, errors.OperationalError):
            return True

        if isinstance(error, errors.InterfaceError):
            if error.msg == 'Use cmd_query_iter for statements with multiple queries.':
                return False
            if error.msg == 'Use multi=True when executing multiple statements':
                return False
            return True

        if isinstance(errors, errors.NotSupportedError):
            return True

        return False

    @classmethod
    def factory(cls, connection_data, session_name=None):
        try:
            instance = cls(connection_data, session_name)
        except mysql.connector.Error as err:
            logger.exception('Connection exception %s', err)
            return (None, err)
        return (instance, None)

    @classmethod
    def test(cls, connection_data):
        connection, err = cls.factory(connection_data)
        if connection is None:
            return (False, str(err))
        connection.close()
        return (True, None)



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from mitzasql.db.model import (DatabasesModel, DBTablesModel, TableModel, QueryModel)

from mitzasql.ui.screens.base_widgets_factory import BaseWidgetsFactory
from mitzasql.ui.widgets.db_view import DBView
from mitzasql.ui.widgets.db_tables_view import DBTablesView
from mitzasql.ui.widgets.table_view import TableView
from mitzasql.ui.widgets.query_view import QueryView

class WidgetsFactory(BaseWidgetsFactory):

    def _get_factory_methods(self):
        factory_methods = {}
        factory_methods['databases_view'] = self.create_databases_view
        factory_methods['db_tables_view'] = self.create_db_tables_view
        factory_methods['table_view'] = self.create_table_view
        factory_methods['query_view'] = self.create_query_view

        return factory_methods

    def create_databases_view(self, connection):
        model = DatabasesModel(connection)
        view = DBView(model, connection)
        self._connect_signal(view, view.SIGNAL_ACTION_QUIT,
                self.change_fsm_state, user_args=['quit'])
        self._connect_signal(view, view.SIGNAL_ACTION_SELECT_DB,
                self.change_fsm_state, user_args=['select'])
        self._connect_signal(view, view.SIGNAL_ACTION_RUN_QUERY,
                self.change_fsm_state, user_args=['run_query'])
        return view

    def create_db_tables_view(self, database, connection):
        model = DBTablesModel(connection, database)
        view = DBTablesView(model, connection)
        self._connect_signal(view, view.SIGNAL_ACTION_QUIT,
                self.change_fsm_state, user_args=['quit'])
        self._connect_signal(view, view.SIGNAL_ACTION_EXIT,
                self.change_fsm_state, user_args=['back'])
        self._connect_signal(view, view.SIGNAL_ACTION_RUN_QUERY,
                self.change_fsm_state, user_args=['run_query'])
        return view

    def create_table_view(self, table, connection):
        model = TableModel(connection, table)
        view = TableView(model, connection)
        self._connect_signal(view, view.SIGNAL_ACTION_QUIT,
                self.change_fsm_state, user_args=['quit'])
        self._connect_signal(view, view.SIGNAL_ACTION_EXIT,
                self.change_fsm_state, user_args=['back'])
        self._connect_signal(view, view.SIGNAL_ACTION_RUN_QUERY,
                self.change_fsm_state, user_args=['run_query'])
        return view

    def create_query_view(self, query, connection):
        model = QueryModel(connection, query)
        view = QueryView(model, connection)
        self._connect_signal(view, view.SIGNAL_ACTION_QUIT,
                self.change_fsm_state, user_args=['quit'])
        self._connect_signal(view, view.SIGNAL_ACTION_EXIT,
                self.change_fsm_state, user_args=['back'])
        self._connect_signal(view, view.SIGNAL_ACTION_RUN_QUERY,
                self.change_fsm_state, user_args=['run_query'])
        return view



# Copyright (c) 2019 Vlad Balmos <vladbalmos@yahoo.com>
# Author: Vlad Balmos <vladbalmos@yahoo.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import urwid

from .emacs_edit import EmacsEdit

class CommandEdit(EmacsEdit):
    '''Implement a basic VIM style command edit widget'''
    SIGNAL_CANCEL = 'cancel_edit'

    def __init__(self, command_processor):
        self._command_processor = command_processor
        self._marker = None
        super().__init__()
        urwid.register_signal(self.__class__, self.SIGNAL_CANCEL)

    def set_marker(self, marker):
        self.set_caption(marker)
        self._marker = marker

    def _show_prev_command(self):
        item = self._command_processor.history.prev
        if not item:
            return
        self.edit_text = item
        self.edit_pos = len(item)

    def _show_next_command(self):
        item = self._command_processor.history.next
        if not item:
            return
        self.edit_text = item
        self.edit_pos = len(item)

    def _cancel_edit(self):
        urwid.emit_signal(self, self.SIGNAL_CANCEL, self)

    def show_last_cmd(self):
        last_cmd = self._command_processor.history.last
        if not last_cmd:
            return
        self.edit_text = last_cmd
        self.edit_pos = len(last_cmd)

    def keypress(self, size, key):
        if (key == 'tab' or key == 'shift tab') and self._marker == ':':
            if key == 'shift tab':
                direction = 'back'
            else:
                direction = 'forward'
            if self._autocomplete(direction):
                return None

        if key != 'tab' and key != 'shift tab':
            self._autocomplete(reset=True)

        if key == 'esc':
            self._cancel_edit()
            return

        key = super().keypress(size, key)
        if key == 'enter':
            text = self.edit_text
            marker = self._marker
            self._cancel_edit()
            self._command_processor.execute(str_cmd_marker=marker,
                    cmd=text)
            return

        # Show previous command
        if key == 'ctrl p':
            self._show_prev_command()
            return

        # Show next command
        if key == 'ctrl n':
            self._show_next_command()
            return

        return key

    def _autocomplete(self, direction='forward', reset=False):
        if reset is True:
            self._command_processor.reset_autocomplete()
            return

        text = self.edit_text
        pos = self.edit_pos
        space_pos = text.find(' ')
        segments = text.split(' ')
        suggestion = None
        new_pos = None

        # Autocomplete first word
        if pos <= len(segments[0]):
            if pos <= space_pos or space_pos == -1:
                suggestion = self._command_processor.autocomplete(segments[0],
                        command=True, direction=direction)

            if suggestion:
                segments[0] = suggestion
                new_pos = len(suggestion)

        # Autocomplete second word
        if pos > space_pos and space_pos > -1:
            next_space_pos = text.find(' ', space_pos + 1)
            if pos > next_space_pos and next_space_pos > -1:
                return

            suggestion = self._command_processor.autocomplete(segments[1],
                    argument=True, command_name=segments[0],
                    direction=direction)
            if suggestion:
                segments[1] = suggestion
                new_pos = len(segments[0]) + len(suggestion) + 1

        if suggestion:
            text = ' '.join(segments)
            self.edit_text = text
            self.edit_pos = new_pos
            return True

    def reset(self):
        self._marker = ''
        self.set_caption('')
        self.edit_text = ''

